//
//  AppCoordinator.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 08.08.23.
//

import UIKit
import KeychainAccess
import AVFoundation
import MBProgressHUD

enum Tab: Int {
  case artists = 0
  case albums = 1
}

class AppCoordinator: Coordinator {

  let window: UIWindow
  let rootViewController: RootViewController
  //TODO: Custom TabBarController?!
  let tabBarController: PinnipedTabBarController
  var childCoordinator: Coordinator?

  let artistsCoordinator: ArtistsCoordinator
  let albumsCoordinator: AlbumsCoordinator

  let fileManager: FileManager
  let apiClient: APIClient
  let keychain: Keychain
  let branding: Branding

  let audioPlayer: AudioPlayer
  let audioDownloader: AudioDownloader
  var currentDownloadTask: Task<Void, Never>?

  init(window: UIWindow, fileManager: FileManager = .default, audioPlayer: AudioPlayer = .shared) {
    self.window = window
    self.fileManager = fileManager
    self.rootViewController = RootViewController()
    self.tabBarController = PinnipedTabBarController()
    self.branding = Branding.shared
    self.audioPlayer = audioPlayer


    keychain = Keychain(service: "pinniped")

    let baseURL: URL?
    if let baseURLString = try? keychain.get(Keychain.Keys.baseURL) {
      baseURL = URL(string: baseURLString)
    } else {
      baseURL = nil
    }

    apiClient = APIClient(baseURL: baseURL, keychain: keychain)

    artistsCoordinator = ArtistsCoordinator(apiClient: apiClient, keychain: keychain)
    albumsCoordinator = AlbumsCoordinator(apiClient: apiClient, keychain: keychain)

    audioDownloader = AudioDownloader(fileManager: fileManager, apiClient: apiClient)
  }

  func start() {

    artistsCoordinator.delegate = self
    albumsCoordinator.delegate = self
    tabBarController.currentSongWidget.delegate = self

    rootViewController.addChild(tabBarController)
    rootViewController.view.addSubview(tabBarController.view)
    tabBarController.didMove(toParent: rootViewController)

    tabBarController.viewControllers = [artistsCoordinator.navigationController, albumsCoordinator.navigationController]

    window.rootViewController = rootViewController
    window.makeKeyAndVisible()

    if try! keychain.get(Keychain.Keys.baseURL) == nil {
      showOnboarding()
    }
    
    artistsCoordinator.start()
    albumsCoordinator.start()
  }

  func showOnboarding() {
    let onboardingCoordinator = OnboardingCoordinator(presentedOn: rootViewController, apiClient: apiClient, keychain: keychain, window: window)
    onboardingCoordinator.delegate = self
    onboardingCoordinator.start()

    childCoordinator = onboardingCoordinator
  }
}

extension AppCoordinator: OnboardingCoordinatorDelegate {
  func onboardingCompleted(_ coordinator: OnboardingCoordinator) {
    coordinator.navigationController.dismiss(animated: true)

    artistsCoordinator.loadArtists(artistsViewController: artistsCoordinator.artistsViewController)
    albumsCoordinator.loadAlbums(albumsViewController: albumsCoordinator.albumsViewController)
  }
}

extension AppCoordinator: AlbumsCoordinatorDelegate {
  func trackSelected(_ coordinator: Coordinator, track: Track, trackList: [Track], album: Album) {

    currentDownloadTask?.cancel()

    audioPlayer.pause()
    tabBarController.currentSongWidget.isHidden = false
    tabBarController.currentSongWidget.isLoading()
    tabBarController.currentSongWidget.update(with: track, from: album)

    let download = Task {
      do {

        try await audioDownloader.downloadAudioFile(for: track)

        _ = await MainActor.run {
          audioPlayer.play(track: track, from: album, trackList: trackList)
          tabBarController.currentSongWidget.play()
        }

      } catch let error as URLError {
        if error.code.rawValue == -999 {
          // do nothing as this request was cancelled
        }
      } catch {
        _ = await MainActor.run {
          tabBarController.currentSongWidget.isHidden = true
        }

        print(error)
      }
    }

    currentDownloadTask = download
  }
}

extension AppCoordinator: ArtistsCoordinatorDelegate {}

extension AppCoordinator: CurrentSongWidgetDelegate {
  func playPauseButtonPressed(_ widget: CurrentSongWidget, isPlaying: Bool) {
    if isPlaying {
      audioPlayer.play()
    } else {
      audioPlayer.pause()
    }
  }
}
