//
//  APIClient.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 08.08.23.
//

import Foundation
import KeychainAccess

class APIClient {

  var baseURL: URL?
  let urlSession: URLSession
  //FIXME: Keychain wirklich hier?
  let keychain: Keychain

  init(baseURL: URL? = nil, urlSession: URLSession = URLSession.shared, keychain: Keychain) {
    self.baseURL = baseURL
    self.urlSession = urlSession
    self.keychain = keychain
  }

  func registerApplication(name: String, redirect_uri: String, scopes: String) async throws -> OAuthAppRegistrationResponse {

    guard let baseURL else { throw FunkwhaleError.noURL}

    let endpoint = FunkwhaleEndpoints.registerApp
    let body = OAuthAppRegistrationRequestBody(name: name, redirect_uris: redirect_uri, scopes: scopes)
    let bodyData = try JSONEncoder().encode(body)
    let request = endpoint.request(baseURL: baseURL, body: bodyData)

    guard let request else { throw FunkwhaleError.noURL }

    let (data, response) = try await urlSession.data(for: request)

    guard let httpResponse = response as? HTTPURLResponse, httpResponse.statusCode == HTTPStatusCode.created.rawValue else {
      throw FunkwhaleError.registrationFailed
    }

    let decoder = JSONDecoder()
    //FIXME: This should be valid??? What's wrong with the API?
    decoder.dateDecodingStrategy = .iso8601

    let registrationResponse = try decoder.decode(OAuthAppRegistrationResponse.self, from: data)

    return registrationResponse
  }

  func getAccessToken(code: String, client_id: String, client_secret: String, redirect_uri: String) async throws -> AccessTokenResponse {

    guard let baseURL else { throw FunkwhaleError.noURL}

    let endpoint = FunkwhaleEndpoints.getToken
    let bodyData = "grant_type=authorization_code&code=\(code)&redirect_uri=\(redirect_uri)&client_id=\(client_id)".data(using: .utf8)
    let request = endpoint.request(baseURL: baseURL, body: bodyData, username: client_id, password: client_secret)

    guard let request else { throw FunkwhaleError.noURL }

    let (data, _) = try await urlSession.data(for: request)
    let decoder = JSONDecoder()

    //TODO: Check for OAuth errors and reasons. For now it will just crash
    let accessToken = try decoder.decode(AccessTokenResponse.self, from: data)

    return accessToken
  }

  func getRefreshToken(refresh_token: String, client_id: String, client_secret: String) async throws -> AccessTokenResponse {

    guard let baseURL else { throw FunkwhaleError.noURL}
    
    let endpoint = FunkwhaleEndpoints.getToken
    let bodyData = "grant_type=refresh_token&refresh_token=\(refresh_token)".data(using: .utf8)
    let request = endpoint.request(baseURL: baseURL, body: bodyData, username: client_id, password: client_secret)

    guard let request else { throw FunkwhaleError.noURL }

    let (data, _) = try await urlSession.data(for: request)
    let decoder = JSONDecoder()

    //TODO: Check for OAuth errors and reasons. For now it will just crash
    let accessToken = try decoder.decode(AccessTokenResponse.self, from: data)

    return accessToken
  }

  func downloadArtists() async throws -> FunkwhaleCollectionResponse<Artist> {
    guard let accessToken = try? keychain.get(Keychain.Keys.accessToken) else { throw FunkwhaleError.artistsDownloadFailed }

    return try await getJSONData(from: .artists, accessToken: accessToken, refreshToken: true)
  }

  func downloadAlbums() async throws -> FunkwhaleCollectionResponse<Album> {
    guard let accessToken = try? keychain.get(Keychain.Keys.accessToken) else { throw FunkwhaleError.albumDownloadFailed }

    return try await getJSONData(from: .albums, accessToken: accessToken, refreshToken: true)
  }

  func downloadAlbums(for artist: Artist) async throws -> FunkwhaleCollectionResponse<Album> {
    guard let accessToken = try? keychain.get(Keychain.Keys.accessToken) else { throw FunkwhaleError.albumDownloadFailed }

    return try await getJSONData(from: .albums, accessToken: accessToken, artistID: artist.id, refreshToken: true)
  }

  func downloadTracks(for album: Album) async throws -> FunkwhaleCollectionResponse<Track> {
    guard let accessToken = try? keychain.get(Keychain.Keys.accessToken) else { throw FunkwhaleError.albumDownloadFailed }

    return try await getJSONData(from: .tracks, accessToken: accessToken, albumID: album.id, refreshToken: true)
  }

  func downloadTrack(for track: Track) async throws -> Data {
    guard let accessToken = try? keychain.get(Keychain.Keys.accessToken) else { throw FunkwhaleError.albumDownloadFailed }

    return try await downloadData(from: .listenToTack(track), accessToken: accessToken)
  }
}

extension APIClient {

  private func downloadData(from endpoint: FunkwhaleEndpoints, accessToken: String, searchString q: String? = nil, artistID: Int? = nil, albumID: Int? = nil, refreshToken: Bool = false) async throws -> Data {
    guard let baseURL else { throw FunkwhaleError.noURL}

    let request = endpoint.request(baseURL: baseURL, token: accessToken)

    guard var request else { throw FunkwhaleError.noURL }
    guard let requestUrl = request.url else {
      throw FunkwhaleError.registrationFailed
    }

    var urlComponents = URLComponents(url: requestUrl, resolvingAgainstBaseURL: true)
    var queryItems = urlComponents?.queryItems ?? []

    if let q {
      let qQueryItem = URLQueryItem(name: "q", value: q)
      queryItems.append(qQueryItem)
    }

    if let artistID {
      let artistQueryItem = URLQueryItem(name: "artist", value: "\(artistID)")
      queryItems.append(artistQueryItem)
    }

    if let albumID {
      let albumQueryItem = URLQueryItem(name: "album", value: "\(albumID)")
      queryItems.append(albumQueryItem)
    }

    urlComponents?.queryItems = queryItems

    request.url = urlComponents?.url

    let (data, response) = try await urlSession.data(for: request)

    guard let httpResponse = response as? HTTPURLResponse else {
      throw FunkwhaleError.albumDownloadFailed
    }

    if refreshToken, httpResponse.statusCode == HTTPStatusCode.unauthorized.rawValue,
       let refreshToken = try keychain.get(Keychain.Keys.refreshToken),
       let clientID = try keychain.get(Keychain.Keys.clientID),
       let clientSecret = try keychain.get(Keychain.Keys.clientSecret) {
      let newTokens = try await getRefreshToken(refresh_token: refreshToken, client_id: clientID, client_secret: clientSecret)

      try keychain.set(newTokens.refresh_token, key: Keychain.Keys.refreshToken)
      try keychain.set(newTokens.access_token, key: Keychain.Keys.accessToken)

      return try await getJSONData(from: endpoint, accessToken: newTokens.access_token, searchString: q, refreshToken: false)
    }

    return data
  }


  private func getJSONData<T: Codable>(from endpoint: FunkwhaleEndpoints, accessToken: String, searchString q: String? = nil, artistID: Int? = nil, albumID: Int? = nil, refreshToken: Bool = false) async throws -> T {

    guard let baseURL else { throw FunkwhaleError.noURL}

    let request = endpoint.request(baseURL: baseURL, token: accessToken)

    guard var request else { throw FunkwhaleError.noURL }
    guard let requestUrl = request.url else {
      throw FunkwhaleError.registrationFailed
    }

    var urlComponents = URLComponents(url: requestUrl, resolvingAgainstBaseURL: true)
    var queryItems = urlComponents?.queryItems ?? []

    if let q {
      let qQueryItem = URLQueryItem(name: "q", value: q)
      queryItems.append(qQueryItem)
    }

    if let artistID {
      let artistQueryItem = URLQueryItem(name: "artist", value: "\(artistID)")
      queryItems.append(artistQueryItem)
    }

    if let albumID {
      let albumQueryItem = URLQueryItem(name: "album", value: "\(albumID)")
      queryItems.append(albumQueryItem)
    }

    urlComponents?.queryItems = queryItems

    request.url = urlComponents?.url

    let (data, response) = try await urlSession.data(for: request)

    guard let httpResponse = response as? HTTPURLResponse else {
      throw FunkwhaleError.albumDownloadFailed
    }

    if refreshToken, httpResponse.statusCode == HTTPStatusCode.unauthorized.rawValue,
       let refreshToken = try keychain.get(Keychain.Keys.refreshToken),
       let clientID = try keychain.get(Keychain.Keys.clientID),
       let clientSecret = try keychain.get(Keychain.Keys.clientSecret) {
      let newTokens = try await getRefreshToken(refresh_token: refreshToken, client_id: clientID, client_secret: clientSecret)

      try keychain.set(newTokens.refresh_token, key: Keychain.Keys.refreshToken)
      try keychain.set(newTokens.access_token, key: Keychain.Keys.accessToken)

      return try await getJSONData(from: endpoint, accessToken: newTokens.access_token, searchString: q, refreshToken: false)
    }

    let decoder = JSONDecoder()

    let dataResponse = try decoder.decode(T.self, from: data)
    return dataResponse
  }

}
