//
//  AlbumDetailsViewController.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 10.08.23.
//

import UIKit

enum AlbumDetailsSection: Hashable {
  case album
  case tracks

  var section: Int {
    switch self {
      case .album: return 0
      case .tracks: return 1
    }
  }
}

enum AlbumDetailsRow: Hashable {
  case album(Album)
  case track(Track)
}

protocol AlbumDetailsViewControllerDelegate: AnyObject {
  func trackSelected(_ viewController: AlbumDetailsViewController, track: Track, trackList: [Track], album: Album)
}

class AlbumDetailsViewController: UIViewController {

  var contentView: AlbumDetailsView {
    view as! AlbumDetailsView
  }

  let tracks: [Track]
  let album: Album

  var dataSource: UITableViewDiffableDataSource<AlbumDetailsSection, AlbumDetailsRow>?
  weak var delegate: AlbumDetailsViewControllerDelegate?

  init(tracks: [Track], album: Album) {

    self.tracks = tracks
    self.album = album

    super.init(nibName: nil, bundle: nil)

    title = album.title

    NotificationCenter.default.addObserver(self, selector: #selector(audioPlaybackStarted(_:)), name: .audioPlaybackStarted, object: nil)
  }

  required init?(coder: NSCoder) { fatalError("init(coder:) has not been implemented") }

  override func loadView() {
    let detailsView = AlbumDetailsView(frame: .zero)

    let dataSource = UITableViewDiffableDataSource<AlbumDetailsSection, AlbumDetailsRow>(tableView: detailsView.tableView) { tableView, indexPath, itemIdentifier in

      //TODO: Consider current track
      switch itemIdentifier {
        case .album(let album):
          guard let cell = tableView.dequeueReusableCell(withIdentifier: AlbumDetailsCoverTableViewCell.reuseIdentifier, for: indexPath) as? AlbumDetailsCoverTableViewCell else { fatalError("Wrong cell WTF?!") }

          cell.configure(with: album)

          return cell

        case .track(let track):
          guard let cell = tableView.dequeueReusableCell(withIdentifier: AlbumDetailsTracksTableViewCell.reuseIdentifier, for: indexPath) as? AlbumDetailsTracksTableViewCell else { fatalError("Wrong cell WTF?!") }

          cell.configure(with: track)

          return cell
      }
    }

    detailsView.tableView.delegate = self
    self.dataSource = dataSource

    view = detailsView
  }

  override func viewDidLoad() {
    super.viewDidLoad()

    var snapshot = NSDiffableDataSourceSnapshot<AlbumDetailsSection, AlbumDetailsRow>()
    snapshot.appendSections([.album, .tracks])
    snapshot.appendItems([.album(album)], toSection: .album)
    snapshot.appendItems(tracks.compactMap { .track($0)}, toSection: .tracks )

    dataSource?.apply(snapshot)

  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)

    if AudioPlayer.shared.isPlaying {
      contentView.tableView.contentInset.bottom = CGFloat.scrollViewBottomInsetForWidget
    }
  }

  //MARK: - Notitifications

  @objc
  func audioPlaybackStarted(_ notification: Notification) {
    if AudioPlayer.shared.isPlaying {
      contentView.tableView.contentInset.bottom = CGFloat.scrollViewBottomInsetForWidget
    }
  }
}

extension AlbumDetailsViewController: UITableViewDelegate {
  func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
    tableView.deselectRow(at: indexPath, animated: true)

    guard indexPath.section == AlbumDetailsSection.tracks.section else { return }

    let track = tracks[indexPath.row]
    delegate?.trackSelected(self, track: track, trackList: tracks, album: album)
  }
}
