//
//  AlbumsCoordinator.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 10.08.23.
//

import UIKit
import KeychainAccess
import MBProgressHUD

protocol AlbumsCoordinatorDelegate: AnyObject {
  func trackSelected(_ coordinator: Coordinator, track: Track, trackList: [Track], album: Album)
}

class AlbumsCoordinator: Coordinator {

  let navigationController: UINavigationController
  let albumsViewController: AlbumsViewController

  let apiClient: APIClient
  let keychain: Keychain

  weak var delegate: AlbumsCoordinatorDelegate?

  init(apiClient: APIClient, keychain: Keychain) {
    navigationController = UINavigationController()
    albumsViewController = AlbumsViewController()

    self.apiClient = apiClient
    self.keychain = keychain
  }

  func start() {
    albumsViewController.delegate = self

    navigationController.navigationBar.prefersLargeTitles = true
    let tabBarItem = UITabBarItem(title: L10n.Albums.title, image: UIImage(systemName: "play.square.stack"), tag: Tab.albums.rawValue)
    tabBarItem.selectedImage = UIImage(systemName: "play.square.stack.fill")
    navigationController.tabBarItem = tabBarItem
    navigationController.pushViewController(albumsViewController, animated: true)
  }

  func loadAlbums(albumsViewController: AlbumsViewController) {
    Task {
      do {
        let albums = try await apiClient.downloadAlbums()
        await MainActor.run {
          albumsViewController.update(with: albums.results.sorted(by: { $0.title.lowercased() < $1.title.lowercased() } ), animated: false)
          albumsViewController.contentView.collectionView.refreshControl?.endRefreshing()
        }
      } catch {
        // do nothing
        await MainActor.run {
          albumsViewController.contentView.collectionView.refreshControl?.endRefreshing()
        }
        print(error.localizedDescription)
      }
    }
  }

}

extension AlbumsCoordinator: AlbumsViewControllerDelegate {
  func showAlbumDetails(_ viewController: UIViewController, album: Album) {
    MBProgressHUD.showAdded(to: navigationController.view, animated: true)

    Task {
      do {

        let tracks = try await apiClient.downloadTracks(for: album)

        _ = await MainActor.run {
          MBProgressHUD.hide(for:navigationController.view, animated: true)

          let albumDetailsViewController = AlbumDetailsViewController(tracks: tracks.results.sorted(by: { $0.position < $1.position}),
                                                                      album: album)
          albumDetailsViewController.delegate = self
          
          navigationController.pushViewController(albumDetailsViewController, animated: true)
        }

      } catch {
        _ = await MainActor.run {
          MBProgressHUD.hide(for:navigationController.view, animated: true)
        }
      }
    }

  }

  func showAccountSettings(_ viewController: AlbumsViewController) {
    //TODO: Implement
  }

  func reloadAlbums(_ viewController: AlbumsViewController) {
    loadAlbums(albumsViewController: viewController)
  }
}

extension AlbumsCoordinator: AlbumDetailsViewControllerDelegate {
  func trackSelected(_ viewController: AlbumDetailsViewController, track: Track, trackList: [Track], album: Album) {
    delegate?.trackSelected(self, track: track, trackList: trackList, album: album)
  }
}
