//
//  ArtistsCoordinator.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 09.08.23.
//

import UIKit
import MBProgressHUD
import KeychainAccess

protocol ArtistsCoordinatorDelegate: AnyObject {
  func trackSelected(_ coordinator: Coordinator, track: Track, trackList: [Track], album: Album)
}

class ArtistsCoordinator: Coordinator {

  let navigationController: UINavigationController
  let artistsViewController: ArtistsViewController

  let apiClient: APIClient
  let keychain: Keychain

  weak var delegate: ArtistsCoordinatorDelegate?

  init(apiClient: APIClient, keychain: Keychain) {
    navigationController = UINavigationController()
    artistsViewController = ArtistsViewController()

    self.apiClient = apiClient
    self.keychain = keychain
  }

  func start() {
    artistsViewController.delegate = self

    navigationController.navigationBar.prefersLargeTitles = true
    navigationController.tabBarItem = UITabBarItem(title: L10n.Artists.title, image: UIImage(systemName: "music.mic"), tag: Tab.artists.rawValue)
    navigationController.pushViewController(artistsViewController, animated: true)

    loadArtists(artistsViewController: artistsViewController)
  }

  func loadArtists(artistsViewController: ArtistsViewController) {
    Task {
      do {
        let artists = try await apiClient.downloadArtists()
        let artistsWithAlbumsSortedByName = artists.results
          .filter { $0.albums.isEmpty == false }
          .sorted(by: { $0.name.lowercased() < $1.name.lowercased() } )
        await MainActor.run {
          artistsViewController.update(with: artistsWithAlbumsSortedByName, animated: false)
          artistsViewController.contentView.tableView.refreshControl?.endRefreshing()
        }
      } catch {
        // do nothing
        print(error.localizedDescription)
      }
    }
  }
}

extension ArtistsCoordinator: ArtistsViewControllerDelegate {

  func showAccountSettings(_ viewController: ArtistsViewController) {
    //TODO: Implement
  }

  func showDetailsForArtist(_ viewController: ArtistsViewController, artist: Artist) {
    MBProgressHUD.showAdded(to: navigationController.view, animated: true)
    Task {
      do {
        let albums = try await apiClient.downloadAlbums(for: artist)
        _ = await MainActor.run {
          MBProgressHUD.hide(for: navigationController.view, animated: true)
          //FIXME: Own ViewController??
          let artistAlbumsViewController = ArtistAlbumsViewController(albums: albums.results, artist: artist)
          artistAlbumsViewController.delegate = self

          self.navigationController.pushViewController(artistAlbumsViewController, animated: true)
        }
      } catch {
        _ = await MainActor.run {
          MBProgressHUD.hide(for: navigationController.view, animated: true)
        }
      }
    }
  }

  func reloadArtists(_ viewController: ArtistsViewController) {
    loadArtists(artistsViewController: viewController)
  }
}

//MARK: - AlbumsViewControllerDelegate
extension ArtistsCoordinator: ArtistAlbumsViewControllerDelegate {
  func showAlbumDetails(_ viewController: UIViewController, album: Album) {
    MBProgressHUD.showAdded(to: navigationController.view, animated: true)

    Task {
      do {

        let tracks = try await apiClient.downloadTracks(for: album)

        _ = await MainActor.run {
          MBProgressHUD.hide(for:navigationController.view, animated: true)

          let albumDetailsViewController = AlbumDetailsViewController(tracks: tracks.results.sorted(by: { $0.position < $1.position}),
                                                                      album: album)
          albumDetailsViewController.delegate = self

          navigationController.pushViewController(albumDetailsViewController, animated: true)
        }

      } catch {
        _ = await MainActor.run {
          MBProgressHUD.hide(for:navigationController.view, animated: true)
        }
      }
    }
  }
}

//MARK: - AlbumDetailsViewControllerDelegate

extension ArtistsCoordinator: AlbumDetailsViewControllerDelegate {
  func trackSelected(_ viewController: AlbumDetailsViewController, track: Track, trackList: [Track], album: Album) {
    delegate?.trackSelected(self, track: track, trackList: trackList, album: album)
  }
}
