//
//  ArtistsViewController.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 09.08.23.
//

import UIKit

protocol ArtistsViewControllerDelegate: AnyObject {
  func showAccountSettings(_ viewController: ArtistsViewController)
  func showDetailsForArtist(_ viewController: ArtistsViewController, artist: Artist)
  func reloadArtists(_ viewController: ArtistsViewController)
}

enum ArtistsSection: Hashable {
  case main
}

enum ArtistsEntry: Hashable {
  case artist(Artist)
}

class ArtistsViewController: UIViewController {

  var contentView: ArtistsView {
    view as! ArtistsView
  }

  var artists: [Artist] = []

  weak var delegate: ArtistsViewControllerDelegate?
  var dataSource: UITableViewDiffableDataSource<ArtistsSection, ArtistsEntry>?

  init() {
    super.init(nibName: nil, bundle: nil)

    title = L10n.Artists.title
    navigationItem.rightBarButtonItem = UIBarButtonItem(image: UIImage(systemName: "person.circle"), style: .done, target: self, action: #selector(ArtistsViewController.showAccountSettings(_:)))

    NotificationCenter.default.addObserver(self, selector: #selector(audioPlaybackStarted(_:)), name: .audioPlaybackStarted, object: nil)
  }

  required init?(coder: NSCoder) { fatalError("init(coder:) has not been implemented") }

  override func loadView() {
    let artistsView = ArtistsView(frame: .zero)

    let dataSource = UITableViewDiffableDataSource<ArtistsSection, ArtistsEntry>(tableView: artistsView.tableView) { tableView, indexPath, itemIdentifier in
      guard let cell = tableView.dequeueReusableCell(withIdentifier: ArtistsTableViewCell.reuseIdentifier, for: indexPath) as? ArtistsTableViewCell else { fatalError("Wrong cell WTF?") }

      switch itemIdentifier {
        case .artist(let artist):
          cell.update(with: artist)
      }

      return cell
    }

    artistsView.tableView.delegate = self
    artistsView.tableView.refreshControl?.addTarget(self, action: #selector(ArtistsViewController.reloadArtists(_:)), for: .valueChanged)

    view = artistsView
    self.dataSource = dataSource
  }

  override func viewDidLoad() {
    super.viewDidLoad()
    delegate?.reloadArtists(self)
  }

  override func viewWillAppear(_ animated: Bool) {
    super.viewWillAppear(animated)

    if AudioPlayer.shared.isPlaying {
      contentView.tableView.contentInset.bottom = CGFloat.scrollViewBottomInsetForWidget
    }
  }

  //MARK: - Notitifications

  @objc
  func audioPlaybackStarted(_ notification: Notification) {
    if AudioPlayer.shared.isPlaying {
      contentView.tableView.contentInset.bottom = CGFloat.scrollViewBottomInsetForWidget
    }
  }

  func update(with artists: [Artist], animated: Bool) {

    self.artists = artists

    var snapshot = NSDiffableDataSourceSnapshot<ArtistsSection, ArtistsEntry>()
    snapshot.appendSections([.main])
    snapshot.appendItems(artists.compactMap { .artist($0) })

    dataSource?.apply(snapshot, animatingDifferences: animated)
  }

  @objc
  func reloadArtists(_ sender: Any) {
    delegate?.reloadArtists(self)
  }

  @objc
  func showAccountSettings(_ sender: Any) {
    delegate?.showAccountSettings(self)
  }
}

extension ArtistsViewController: UITableViewDelegate {
  func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
    tableView.deselectRow(at: indexPath, animated: true)

    let artist = artists[indexPath.row]
    delegate?.showDetailsForArtist(self, artist: artist)
  }
}
