//
//  CurrentSongWidget.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 22.08.23.
//

import UIKit
import Kingfisher

protocol CurrentSongWidgetDelegate: AnyObject {
  func playPauseButtonPressed(_ widget: CurrentSongWidget, isPlaying: Bool)
}

class CurrentSongWidget: UIView {
  let coverImageView: UIImageView
  let titleLabel: UILabel
  let playPauseButton: UIButton

  private let contentStackView: UIStackView
  var songIsPlaying = false

  var delegate: CurrentSongWidgetDelegate?

  override init(frame: CGRect) {
    //TODO: Add tap to see the fullscreen-screen with the current song
    coverImageView = UIImageView()
    coverImageView.translatesAutoresizingMaskIntoConstraints = false
    coverImageView.image = Asset.albumPlaceholder.image
    coverImageView.layer.cornerRadius = 8.0
    coverImageView.layer.maskedCorners = [.layerMaxXMaxYCorner, .layerMaxXMinYCorner, .layerMinXMaxYCorner, .layerMinXMinYCorner]
    coverImageView.layer.masksToBounds = true

    titleLabel = UILabel()
    titleLabel.translatesAutoresizingMaskIntoConstraints = false
    titleLabel.numberOfLines = 1
    titleLabel.textColor = .white
    titleLabel.font = UIFont.preferredFont(forTextStyle: .headline)

    var playPauseButtonConfiguration = UIButton.Configuration.plain()
    playPauseButtonConfiguration.baseForegroundColor = .white
    playPauseButtonConfiguration.activityIndicatorColorTransformer = UIConfigurationColorTransformer({ _ in
      return UIColor.white
    })
    playPauseButton = UIButton(configuration: playPauseButtonConfiguration)

    contentStackView = UIStackView(arrangedSubviews: [coverImageView, titleLabel, playPauseButton])
    contentStackView.translatesAutoresizingMaskIntoConstraints = false
    contentStackView.alignment = .center
    contentStackView.axis = .horizontal
    contentStackView.spacing = 15
    playPauseButton.translatesAutoresizingMaskIntoConstraints = false

    super.init(frame: frame)

    let blur = UIBlurEffect(style: .systemMaterialDark)
    let blurEffectView = UIVisualEffectView(effect: blur)
    blurEffectView.frame = bounds
    blurEffectView.layer.cornerRadius = 8
    blurEffectView.layer.masksToBounds = true
    blurEffectView.autoresizingMask = [.flexibleWidth, .flexibleHeight]
    addSubview(blurEffectView)

    //TODO: Add additional spacing to tableViews below

    addSubview(contentStackView)
    layer.cornerRadius = 8
    setupConstraints()

    playPauseButton.addTarget(self, action: #selector(playPauseButtonPressed(_:)), for: .touchUpInside)
  }

  required init?(coder: NSCoder) { fatalError("init(coder:) has not been implemented") }

  private func setupConstraints() {
    let constraints = [
      contentStackView.topAnchor.constraint(equalTo: topAnchor, constant: 8),
      contentStackView.leadingAnchor.constraint(equalTo: leadingAnchor, constant: 8),
      trailingAnchor.constraint(equalTo: contentStackView.trailingAnchor, constant: 8),
      bottomAnchor.constraint(equalTo: contentStackView.bottomAnchor, constant: 8),

      coverImageView.widthAnchor.constraint(equalToConstant: 44),
      coverImageView.heightAnchor.constraint(equalTo: coverImageView.widthAnchor),
      playPauseButton.widthAnchor.constraint(equalToConstant: 40),
    ]

    NSLayoutConstraint.activate(constraints)
  }

  //MARK: - Actions

  @objc
  func playPauseButtonPressed(_ sender: UIButton) {
    if songIsPlaying {
      pause()
    } else {
      play()
    }

    delegate?.playPauseButtonPressed(self, isPlaying: songIsPlaying)
  }

  func isLoading() {
    playPauseButton.configuration?.showsActivityIndicator = true
  }

  func update(with track: Track, from album: Album) {
    titleLabel.text = track.title

    if let imageURL = album.cover?.urls.medium_square_crop {
      coverImageView.kf.setImage(with: URL(string: imageURL))
    }
  }

  func play() {
    songIsPlaying = true
    playPauseButton.configuration?.showsActivityIndicator = false
    playPauseButton.setImage(UIImage(systemName: "pause.fill"), for: .normal)
  }

  func pause() {
    songIsPlaying = false
    playPauseButton.configuration?.showsActivityIndicator = false
    playPauseButton.setImage(UIImage(systemName: "play.fill"), for: .normal)
  }
}
