//
//  OnboardingCoordinator.swift
//  pinniped-ios
//
//  Created by Nathan Mattes on 08.08.23.
//

import UIKit
import MBProgressHUD
import KeychainAccess
import AuthenticationServices

protocol OnboardingCoordinatorDelegate: AnyObject {
  func onboardingCompleted(_ coordinator: OnboardingCoordinator)
}

class OnboardingCoordinator: NSObject, Coordinator {

  let navigationController: UINavigationController
  let welcomeViewController: WelcomeViewController
  let presentedOn: UIViewController
  let apiClient: APIClient
  let keychain: Keychain
  let window: UIWindow

  weak var delegate: OnboardingCoordinatorDelegate?

  init(presentedOn: UIViewController, apiClient: APIClient, keychain: Keychain, window: UIWindow) {
    
    navigationController = UINavigationController()
    navigationController.navigationBar.prefersLargeTitles = true
    navigationController.modalPresentationStyle = .overFullScreen

    welcomeViewController = WelcomeViewController()

    self.presentedOn = presentedOn
    self.apiClient = apiClient
    self.keychain = keychain
    self.window = window
  }

  func start() {
    welcomeViewController.delegate = self

    navigationController.viewControllers = [welcomeViewController]

    presentedOn.present(navigationController, animated: true)
  }
}

//MARK: - WelcomeViewControllerDelegate
extension OnboardingCoordinator: WelcomeViewControllerDelegate {
  func connectFunkwhale(_ viewController: WelcomeViewController) {
    let connectToFunkwhaleViewController = ConnectToFunkwhaleViewController()
    connectToFunkwhaleViewController.delegate = self

    navigationController.pushViewController(connectToFunkwhaleViewController, animated: true)
  }

  func subscribe(_ viewController: WelcomeViewController) {
    //TODO: Implement.
    delegate?.onboardingCompleted(self)
  }
}

//MARK:
extension OnboardingCoordinator: ConnectToFunkwhaleViewControllerDelegate {
  func login(_ viewController: ConnectToFunkwhaleViewController, urlString: String) {
    let baseURL: URL?
    if urlString.hasPrefix("https://") {
      baseURL = URL(string: urlString)
    } else {
      baseURL = URL(string: "https://\(urlString)")
    }

    guard let baseURL else { return }

    MBProgressHUD.showAdded(to: navigationController.view, animated: true)

    Task {
      do {

        let redirectURI = "pinniped-ios://oauth2"
        let name = "pinniped for iOS"
        apiClient.baseURL = baseURL

        let response = try await apiClient.registerApplication(name: name, redirect_uri: redirectURI, scopes: "read write")

        let client_id = response.client_id
        let client_secret = response.client_secret

        try keychain.set(client_id, key: Keychain.Keys.clientID)
        try keychain.set(client_secret, key: Keychain.Keys.clientSecret)

        _ = await MainActor.run {
          MBProgressHUD.hide(for: navigationController.view, animated:true)
        }

        guard let url = FunkwhaleEndpoints.authorize.url(baseURL: baseURL) else { throw FunkwhaleError.loginFailed }

        var urlComponents = URLComponents(url: url, resolvingAgainstBaseURL: false)

        urlComponents?.queryItems = [
          URLQueryItem(name: "client_id", value: client_id),
          URLQueryItem(name: "response_type", value: "code"),
          URLQueryItem(name: "scope", value: "read write"),
          URLQueryItem(name: "redirect_uri", value: redirectURI)
        ]

        guard let authUrl = urlComponents?.url else {
          fatalError("invalid url")
        }

        let session = ASWebAuthenticationSession(url: authUrl, callbackURLScheme: "pinniped-ios") { callbackURL, error in

          guard error == nil, let callbackURL = callbackURL,
                let queryItems = URLComponents(string: callbackURL.absoluteString)?.queryItems,
                let code = queryItems.first(where: { $0.name == "code" })?.value
          else {
            print("An error occurred when attempting to sign in: \(error?.localizedDescription ?? "(No Details)")")
            return
          }

          Task {
            do {
              let token = try await self.apiClient.getAccessToken(code: code, client_id: client_id, client_secret: client_secret, redirect_uri: redirectURI)

              try self.keychain.set(token.access_token, key: Keychain.Keys.accessToken)
              try self.keychain.set(token.refresh_token, key: Keychain.Keys.refreshToken)
              try self.keychain.set(baseURL.absoluteString, key: Keychain.Keys.baseURL)

              _ = await MainActor.run {
                self.welcomeViewController.loginSuccessful()
                self.navigationController.popToRootViewController(animated:true)
              }
            } catch {
              print(error.localizedDescription)
            }
          }

        }
        session.presentationContextProvider = self
        session.start()
      } catch {
        _ = await MainActor.run {
          MBProgressHUD.hide(for: navigationController.view, animated:true)
          //TODO: Show alert with error
        }
      }
    }
  }
}

extension OnboardingCoordinator: ASWebAuthenticationPresentationContextProviding {
  func presentationAnchor(for session: ASWebAuthenticationSession) -> ASPresentationAnchor {
    return window
  }
}
